import os
import io
import glob
from typing import List, Tuple, Optional, Union

import cv2
import numpy as np
import imagehash
import requests
import PIL
from PIL import Image
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from tensorflow.keras.preprocessing import image as keras_image

from pengu.utils.function import alpha_numeric_sorted
from pengu.exceptions import DownloadError


def pil2cv(img: Image.Image) -> np.ndarray:
    """Convert PIL.Image.Image to np.ndarray(opencv)
    Args:
        img (Image.Image): PIL image
    Returns:
        np.ndarray: converted image array (opencv)
    """
    img_np = np.array(img, dtype=np.uint8)
    if img_np.ndim == 2:  # L
        pass
    elif img_np.shape[2] == 3:  # RGB
        img_np = cv2.cvtColor(img_np, cv2.COLOR_RGB2BGR)
    elif img_np.shape[2] == 4:  # RGBA
        img_np = cv2.cvtColor(img_np, cv2.COLOR_RGBA2BGRA)
    return img_np


def cv2pil(img: np.ndarray) -> Image.Image:
    """Convert np.ndarray(opencv) to PIL.Image.Image
    Args:
        img (np.ndarray): OpenCV image array
    Returns:
        Image.Image: PIL image
    """
    img_np = img.copy()
    if img_np.ndim == 2:  # L
        pass
    elif img_np.shape[2] == 3:  # RGB
        img_np = img_np[:, :, ::-1]
    elif img_np.shape[2] == 4:  # RGBA
        img_np = img_np[:, :, [2, 1, 0, 3]]
    img_pil = Image.fromarray(img_np)
    return img_pil


def download_image(url: str,
                   timeout: float = 1.0,
                   retry_count: int = 3,
                   retry_backoff_factor: float = 0.5) -> Image.Image:
    """Download Image from image url

    Args:
        url (str): image url

    Raises:
        DownloadError

    Returns:
        Image.Image: PIL Image.
    """
    session = requests.Session()

    retries = Retry(total=retry_count,
                    backoff_factor=retry_backoff_factor)

    session.mount('http://', HTTPAdapter(max_retries=retries))
    session.mount('https://', HTTPAdapter(max_retries=retries))

    response = session.get(url, timeout=timeout)

    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise DownloadError(f"Failed to download Image. url: {url}") from e

    try:
        img = Image.open(io.BytesIO(response.content))
    except OSError as e:
        raise DownloadError(f"Failed to open image downloaded. url: {url}") from e

    return img


def load_image(file_path: str,
               target_size: Tuple[int, int]) -> np.ndarray:
    """load image from file path.

    Args:
        file_path_list (List[str]): Image file path list.
        target_size (Tuple[int, int]): Size when load image.

    Returns:
        np.ndarray: Preprocessed and resized image array. Shape is (h, w, ch)

    """
    img = keras_image.load_img(file_path, target_size=target_size)
    img = np.asarray(img)

    return img


def list_pictures(dir_path: str,
                  n_use: int = -1,
                  ext: Tuple[str, ...] = ('jpg', 'jpeg', 'bmp', 'png', 'tif')) -> List[str]:
    """Lists all pictures in a directory, not including all subdirectories.

    Args
        dir_path (str): Absolute path to the directory.
        ext (Tuple[str, ...]): Extensions of the pictures.

    Returns
        A list of paths.

    """
    ext = tuple('.%s' % e for e in ((ext,) if isinstance(ext, str) else ext))

    picture_file_list = [file for file in glob.glob(os.path.join(dir_path, '*'))
                         if file.lower().endswith(ext)]

    picture_file_list = alpha_numeric_sorted(picture_file_list)
    if n_use != -1:
        picture_file_list = picture_file_list[:n_use]

    return picture_file_list


def get_image_meta(data: Union[np.ndarray, Image.Image, str],
                   data_format: str,
                   hash_method: str) -> Tuple[Tuple[int, int],
                                              Optional[str],
                                              Optional[imagehash.ImageHash],
                                              Optional[str]]:
    """Get image meta data(size[w, h], mode, hash)

    Args:
        data (Union[np.ndarray, Image.Image, str]):
            image data. support data_format data.
        data_format (str):
            support "url" and "file"(path), "pil", "opencv".
        hash_method (str):
            support "phash" and "average", "dhash", "whash".

    Returns:
        Tuple[Tuple[int, int], Optional[str], Optional[imagehash.ImageHash]]]:
            (size[w, h], image_mode, hash, img_format).
            if DownloadError, ((0,0), None, None, None).
    """
    if isinstance(data, str):
        if data_format == 'url':
            try:
                img: Image.Image = download_image(data)
            except DownloadError:
                return (0, 0), None, None, None
        elif data_format == 'file':
            img = Image.open(data)
        else:
            raise ValueError(f"Invalid data: {data}, data_format: {data_format}")
    elif type(data).__module__.startswith(PIL.__name__):
        if data_format == 'pil':
            img = data
        else:
            raise ValueError(f"Invalid data. data_format: {data_format}")
    elif type(data).__module__.startswith(np.__name__):
        if data_format == 'opencv':
            img = cv2pil(data)
        else:
            raise ValueError(f"Invalid data. data_format: {data_format}")
    else:
        raise ValueError("Invalid data. data_type: "
                         f"{type(data).__module__}, data_format: {data_format}")

    if hash_method == "phash":
        hash_func = imagehash.phash
    elif hash_method == "average":
        hash_func = imagehash.average_hash
    elif hash_method == "dhash":
        hash_func = imagehash.dhash
    elif hash_method == "whash":
        hash_func = imagehash.whash
    else:
        raise ValueError(f"Invalid hashmethod. hash_method: {hash_method}")

    img_hash = hash_func(img)
    img_format = img.format
    return img.size, img.mode, img_hash, img_format



from typing import Optional, Tuple, Dict, Any

import tensorflow as tf

AUTOTUNE = tf.data.experimental.AUTOTUNE


def prepare_tfds(ds: tf.data.Dataset,
                 batch_size: int,
                 shuffle: bool = False,
                 preprocess: Optional[tf.keras.Sequential] = None,
                 augment: Optional[tf.keras.Sequential] = None) -> tf.data.Dataset:
    if preprocess is not None:
        ds = ds.map(lambda x, y: (preprocess(x), y),
                    num_parallel_calls=AUTOTUNE)

    ds = ds.cache()
    if shuffle:
        ds = ds.shuffle(batch_size * 8)

    ds = ds.batch(batch_size)

    if augment is not None:
        ds = ds.map(lambda x, y: (augment(x, training=True), y),
                    num_parallel_calls=AUTOTUNE)
    ds = ds.prefetch(buffer_size=AUTOTUNE)
    return ds


def build_preprocess(rescale_scale: str,
                     rescale_offset: float,
                     img_size: Optional[Tuple[int, int]]) -> tf.keras.Sequential:
    preprocess_list = []
    if img_size is not None:
        preprocess_list.append(
            tf.keras.layers.experimental.preprocessing.Resizing(img_size[0], img_size[1])
        )

    rescale_scale_float = eval(rescale_scale)
    preprocess_list.append(
        tf.keras.layers.experimental.preprocessing.Rescaling(
            rescale_scale_float, offset=rescale_offset)
    )
    return tf.keras.Sequential(preprocess_list)


def build_augment(augment_params: Dict[str, Any]) -> tf.keras.Sequential:
    augment_list = []
    for key, kwargs in augment_params.items():
        augment_class = getattr(tf.keras.layers.experimental.preprocessing, key)
        augment_list.append(augment_class(**kwargs))
    return tf.keras.Sequential(augment_list)


def build_optimizer(name: str, params: Dict[str, Any] = {}) -> tf.keras.optimizers.Optimizer:
    opt_class = getattr(tf.keras.optimizers, name)
    return opt_class(**params)



import argparse
from dataclasses import dataclass, asdict
from pathlib import Path
from typing import List, Tuple, Optional, Dict

from pengu.crawler import ImagesDataCSV
from pengu.utils.file_io import YamlConfig, save_yaml
from pengu.mlflow import MlflowArtifactsConfig, MlflowConfig, MlflowWriter
from pengu.dataset.image import ImageTFRecord


@dataclass
class PreprocessImagesConfig(YamlConfig):
    input_data_csv: MlflowArtifactsConfig
    input_images: MlflowArtifactsConfig
    ratio: List[float]
    img_size: Tuple[int, int] = (500, 500)
    label2index: Optional[Dict[str, int]] = None


@dataclass
class ProprocessImagesMainConfig(YamlConfig):
    preprocess: PreprocessImagesConfig
    mlflow: MlflowConfig


def main():
    parser = argparse.ArgumentParser(
        description="Remove images with duplicate hash and "
                    "Split dataframe to train and val, test dataframe")
    parser.add_argument("--config_path", "-config",
                        type=Path, metavar="path", required=True,
                        help="Config(.yml) file path")
    parser.add_argument("--output_preprocessed_csv_dir", "-csv", type=Path, metavar="Path",
                        required=True,
                        help="Path to result path")
    args = parser.parse_args()

    config: ProprocessImagesMainConfig = ProprocessImagesMainConfig.load(args.config_path)
    preprocess_config: PreprocessImagesConfig = config.preprocess
    mlflow_config: MlflowConfig = config.mlflow

    # MlflowWriter
    writer = MlflowWriter(experiment_name=mlflow_config.experiment_name,
                          tracking_uri=mlflow_config.tracking_uri,
                          registry_uri=mlflow_config.registry_uri)
    writer.log_params_from_config(preprocess_config)
    # download artifacts from Mlflow
    input_csv_config: MlflowArtifactsConfig = preprocess_config.input_data_csv
    image_data_csv_path = writer.download_artifacts(
        run_id=input_csv_config.run_id,
        path=input_csv_config.path,
        dst_path=args.output_preprocessed_csv_dir
    )
    input_images_config: MlflowArtifactsConfig = preprocess_config.input_images
    images_path = writer.download_artifacts(
        run_id=input_images_config.run_id,
        path=input_images_config.path,
        dst_path=args.output_preprocessed_csv_dir
    )

    # preprocess image data csv
    image_data_csv = ImagesDataCSV(image_data_csv_path)
    image_data_csv.drop_duplicates_hash()
    ratio: Tuple[float, float, float] = tuple(preprocess_config.ratio)
    train_df, val_df, test_df = image_data_csv.split_train_val_test(
        ratio=ratio)
    train_csv_path = args.output_preprocessed_csv_dir / "train.csv"
    val_csv_path = args.output_preprocessed_csv_dir / "val.csv"
    test_csv_path = args.output_preprocessed_csv_dir / "test.csv"
    train_df.to_csv(train_csv_path, index=False)
    val_df.to_csv(val_csv_path, index=False)
    test_df.to_csv(test_csv_path, index=False)

    # make tf.data.Dataset
    train_tfrec_path = args.output_preprocessed_csv_dir / "train.tfrec"
    val_tfrec_path = args.output_preprocessed_csv_dir / "val.tfrec"
    test_tfrec_path = args.output_preprocessed_csv_dir / "test.tfrec"
    if preprocess_config.label2index is None:
        label2index = None
    else:
        label2index = asdict(preprocess_config.label2index)
    label2index_path = args.output_preprocessed_csv_dir / "label2index.yaml"
    train_label2index: Dict[str, int] = ImageTFRecord.write_tfrecord(
        df=train_df,
        img_dst_path=images_path,
        tfrecord_path=train_tfrec_path,
        label2index=label2index,
        img_size=preprocess_config.img_size)
    ImageTFRecord.write_tfrecord(df=val_df,
                                 img_dst_path=images_path,
                                 tfrecord_path=val_tfrec_path,
                                 label2index=train_label2index,
                                 img_size=preprocess_config.img_size)
    ImageTFRecord.write_tfrecord(df=test_df,
                                 img_dst_path=images_path,
                                 tfrecord_path=test_tfrec_path,
                                 label2index=train_label2index,
                                 img_size=preprocess_config.img_size)
    save_yaml(train_label2index, label2index_path)

    # add artifacts
    writer.log_artifact(train_csv_path)
    writer.log_artifact(val_csv_path)
    writer.log_artifact(test_csv_path)
    writer.log_artifact(label2index_path)
    writer.log_artifact(train_tfrec_path)
    writer.log_artifact(val_tfrec_path)
    writer.log_artifact(test_tfrec_path)
    writer.set_terminated()


if __name__ == "__main__":
    main()



import json
import inspect
import pickle
from dataclasses import dataclass, asdict
from pathlib import Path
from typing import Any, Optional

import yaml


def save_pkl(data: Any, path: Path) -> None:
    """Save data to pickle file path

    Args:
        data (Any): Object to save
        path (Path): pickle file path to save

    """
    with open(path, mode='wb') as f:
        pickle.dump(data, f)


def load_pkl(path: Path) -> Any:
    """Load pickle file path

    Args:
        path (Path): pickle file path to load

    Returns:
        Any: Object that loaded the pickle file path

    """
    with open(path, mode='rb') as f:
        out = pickle.load(f)
    return out


def save_json(data: Any, path: Path) -> None:
    """Save data to json file path

    Args:
        data (Any): Data to save
        path (Path): Json file path to load

    """
    with open(path, 'w') as f:
        json.dump(data, f)


def load_json(path: Path) -> dict:
    """Load json file path

    Args:
        path (Path): Json file path to load

    Returns:
        dict: Data read from json file

    """
    with open(path) as f:
        json_dict = json.load(f)
    return json_dict


def save_yaml(data: Any, path: Path) -> None:
    """Save data to yaml file path

    Args:
        data (Any): Data to save
        path (Path): Yaml file path to save

    """
    with open(path, 'w') as f:
        yaml.dump(data, f, default_flow_style=False)


def load_yaml(path: Path) -> dict:
    """Load yaml file path

    Args:
        path (Path): Yaml file path to load

    Returns:
        dict: Data read from yaml file

    """
    with open(path) as f:
        yaml_dict = yaml.load(f, Loader=yaml.FullLoader)
    return yaml_dict


@dataclass
class YamlConfig:
    def save(self, config_path: Path):
        """Export config as YAML file """
        config_path.parent.mkdir(exist_ok=True)

        def convert_dict_for_pathlib(data):
            for key, val in data.items():
                if isinstance(val, Path):
                    data[key] = str(val)
                if isinstance(val, dict):
                    data[key] = convert_dict_for_pathlib(val)
            return data

        with open(config_path, 'w') as f:
            yaml.dump(convert_dict_for_pathlib(asdict(self)), f)

    @classmethod
    def load(cls, config_path: Path):
        """Load config from YAML file """

        def convert_from_dict(parent_cls, data):
            for key, val in data.items():
                child_class = parent_cls.__dataclass_fields__[key].type
                if child_class == Optional[Path] and val is not None:
                    data[key] = Path(val)
                if child_class == Path:
                    data[key] = Path(val)
                if hasattr(child_class, "_name"):
                    if child_class._name == "Tuple":
                        data[key] = tuple(val)
                if inspect.isclass(child_class) and issubclass(child_class, YamlConfig):
                    data[key] = child_class(**convert_from_dict(child_class, val))
            return data

        with open(config_path) as f:
            config_data = yaml.full_load(f)
            # recursively convert config item to YamlConfig
            config_data = convert_from_dict(cls, config_data)

        return cls(**config_data)  # type: ignore



import argparse
from pathlib import Path
from dataclasses import dataclass
from typing import List

from icrawler.builtin import GoogleImageCrawler

from pengu.crawler import ImagesDataCSV, SaveResultImageDownloader, FixedGoogleParser
from pengu.utils.file_io import YamlConfig
from pengu.mlflow import MlflowWriter, MlflowConfig


@dataclass
class CrawlerConfig(YamlConfig):
    keywords: List[str]
    labels: List[str]
    start_date: List[int]  # [2020, 1, 1]
    end_date: List[int]


@dataclass
class CrawlerMainConfig(YamlConfig):
    crawler: CrawlerConfig
    mlflow: MlflowConfig


def main():
    # parse args
    parser = argparse.ArgumentParser(description="crawling images from google image search")
    parser.add_argument("--config_path", "-config",
                        type=Path, metavar="path", required=True,
                        help="Config(.yml) file path")
    parser.add_argument("--output_result_csv", "-csv", type=Path, metavar="Path",
                        required=True,
                        help="Path to result path")
    parser.add_argument("--output_images_root_dir", "-i", type=Path, metavar="Path",
                        required=True,
                        help="Path to the root directory for images to save")
    parser.add_argument("--max_num", "-n", type=int, metavar="crawler",
                        required=True,
                        help="Downloaded image number has reached required number")
    parser.add_argument("--feeder_threads", "-f_t", type=int, metavar="crawler", default=1,
                        help="thread number used by feeder")
    parser.add_argument("--parser_threads", "-p_t", type=int, metavar="crawler", default=4,
                        help="thread number used by parser")
    parser.add_argument("--downloader_threads", "-d_t", type=int, metavar="crawler", default=5,
                        help="thread number used by downloader")
    args = parser.parse_args()

    config: CrawlerMainConfig = CrawlerMainConfig.load(args.config_path)
    crawler_config: CrawlerConfig = config.crawler
    mlflow_config: MlflowConfig = config.mlflow

    # MlflowWriter
    writer = MlflowWriter(experiment_name=mlflow_config.experiment_name,
                          tracking_uri=mlflow_config.tracking_uri,
                          registry_uri=mlflow_config.registry_uri)
    writer.log_params_from_config(crawler_config)

    # write header
    if not args.output_result_csv.is_file():
        ImagesDataCSV.write_header(args.output_result_csv)

    filters = dict(
        size="large",
        license="commercial,modify",
        date=(tuple(crawler_config.start_date), tuple(crawler_config.end_date))
    )

    for keyword, label in zip(crawler_config.keywords, crawler_config.labels):
        images_root_dir = args.output_images_root_dir / keyword.replace(" ", "_")
        crawler = GoogleImageCrawler(
            feeder_threads=args.feeder_threads,
            parser_threads=args.parser_threads,
            downloader_threads=args.downloader_threads,
            storage={"root_dir": images_root_dir},
            downloader_cls=SaveResultImageDownloader,
            parser_cls=FixedGoogleParser,
            extra_downloader_args={"image_data_csv_path": args.output_result_csv,
                                   "label": label}
        )
        crawler.crawl(keyword=keyword, filters=filters, max_num=args.max_num)

    # add artifacts
    writer.log_artifact(args.output_result_csv)
    writer.log_artifacts(args.output_images_root_dir)
    writer.set_terminated()


if __name__ == "__main__":
    main()



import argparse
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple, Dict, Any

import tensorflow as tf

from pengu.dataset.image import ImageTFRecord
from pengu.utils.file_io import YamlConfig, load_yaml
from pengu.mlflow import MlflowConfig, MlflowWriter, MlflowCallback
from pengu.model.utils import ModelConfig, build_imagenet_model
from pengu.model.train_utils import prepare_tfds, build_preprocess, build_augment, build_optimizer


AUTOTUNE = tf.data.experimental.AUTOTUNE


@dataclass
class MlflowPreprocessedArtifactsConfig(YamlConfig):
    run_id: str
    train_path: Path
    val_path: Path
    test_path: Path
    label2index_path: Path


@dataclass
class PreprocessConfig(YamlConfig):
    rescale_scale: str  # eval(resize_scale). Exaple: "1./127.5"
    rescale_offset: float
    img_size: Tuple[int, int]


@dataclass
class OptimizerConfig(YamlConfig):
    name: str
    params: Dict[str, Any]


@dataclass
class TrainConfig(YamlConfig):
    input_preprocessed_data: MlflowPreprocessedArtifactsConfig
    preprocess: PreprocessConfig
    augment: Dict[str, dict]  # tf.keras.layers.experimental.preprocessing.{key}(**values)
    model: ModelConfig
    optimizer: OptimizerConfig


@dataclass
class TrainMainConfig(YamlConfig):
    train: TrainConfig
    mlflow: MlflowConfig


def main():
    parser = argparse.ArgumentParser(
        description="Train Keras Model")
    parser.add_argument("--config_path", "-config",
                        type=Path, metavar="path", required=True,
                        help="Config(.yml) file path")
    parser.add_argument("--input_preprocessed_dir", "-i", type=Path, metavar="Path",
                        required=True,
                        help="Path to save preprocessed data")
    parser.add_argument("--output_model_path", "-m", type=Path, metavar="Path",
                        required=True,
                        help="Path to save model")
    parser.add_argument("--tb_logs_path", "-t", type=Path, metavar="Path",
                        required=True,
                        help="Path to write logs for tensorboard")
    parser.add_argument("--batch_size", "-bt", type=int, metavar="train", default=32,
                        help="Batch size")
    parser.add_argument("--epochs", "-e", type=int, metavar="train", default=3,
                        help="epoch")
    args = parser.parse_args()

    config: TrainMainConfig = TrainMainConfig.load(args.config_path)
    train_config: TrainConfig = config.train
    mlflow_config: MlflowConfig = config.mlflow

    # MlflowWriter
    writer = MlflowWriter(experiment_name=mlflow_config.experiment_name,
                          tracking_uri=mlflow_config.tracking_uri,
                          registry_uri=mlflow_config.registry_uri)
    writer.log_params_from_config(train_config)
    # download artifacts from Mlflow
    input_config = train_config.input_preprocessed_data
    train_tfrecord_path = writer.download_artifacts(
        run_id=input_config.run_id,
        path=str(input_config.train_path),
        dst_path=args.input_preprocessed_dir
    )
    val_tfrecord_path = writer.download_artifacts(
        run_id=input_config.run_id,
        path=str(input_config.val_path),
        dst_path=args.input_preprocessed_dir
    )
    test_tfrecord_path = writer.download_artifacts(
        run_id=input_config.run_id,
        path=str(input_config.test_path),
        dst_path=args.input_preprocessed_dir
    )
    label2index_path = writer.download_artifacts(
        run_id=input_config.run_id,
        path=str(input_config.label2index_path),
        dst_path=args.input_preprocessed_dir
    )
    train_ds = ImageTFRecord.read_tfrecord(train_tfrecord_path)
    val_ds = ImageTFRecord.read_tfrecord(val_tfrecord_path)
    test_ds = ImageTFRecord.read_tfrecord(test_tfrecord_path)
    label2index: Dict[str, int] = load_yaml(label2index_path)

    # build preprocess
    preprocess_config = train_config.preprocess
    preprocess = build_preprocess(rescale_scale=preprocess_config.rescale_scale,
                                  rescale_offset=preprocess_config.rescale_offset,
                                  img_size=preprocess_config.img_size)
    # build augment
    augment_config = train_config.augment
    augment = build_augment(augment_config)
    # prepare tf.data.Dataset
    train_ds = prepare_tfds(
        ds=train_ds,
        batch_size=args.batch_size,
        shuffle=True,
        preprocess=preprocess,
        augment=augment
    )
    val_ds = prepare_tfds(
        ds=val_ds,
        batch_size=args.batch_size,
        shuffle=True,
        preprocess=preprocess,
        augment=augment
    )
    test_ds = prepare_tfds(
        ds=test_ds,
        batch_size=args.batch_size,
        shuffle=True,
        preprocess=preprocess,
        augment=augment
    )
    # build_optimizer
    optimizer_config = train_config.optimizer
    optimizer = build_optimizer(optimizer_config.name, optimizer_config.params)
    # build keras model
    model_config = train_config.model
    model_params = model_config.params if model_config.params is not None else {}
    img_shape = preprocess_config.img_size + (3,)
    base_model = build_imagenet_model(
        model_config.name,
        input_shape=img_shape,
        include_top=False,
        **model_params)
    base_model.trainable = False
    model = tf.keras.Sequential([
        base_model,
        tf.keras.layers.GlobalAveragePooling2D(),
        tf.keras.layers.Dropout(0.2),
        tf.keras.layers.Dense(len(label2index))
    ])
    model.compile(optimizer=optimizer,
                  loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
                  metrics=['accuracy'])
    # build callbacks
    args.output_model_path.parent.mkdir(parents=True, exist_ok=True)
    args.tb_logs_path.mkdir(parents=True, exist_ok=True)
    callbacks = [
        tf.keras.callbacks.EarlyStopping(monitor="val_loss", patience=3),
        tf.keras.callbacks.ModelCheckpoint(args.output_model_path),
        tf.keras.callbacks.TensorBoard(args.tb_logs_path),
        MlflowCallback(writer=writer)
    ]
    # train model
    model.fit(
        train_ds,
        validation_data=val_ds,
        epochs=args.epochs,
        callbacks=callbacks
    )

    # evaluate model
    loss, accuracy = model.evaluate(test_ds)
    writer.log_metric(key="test_loss", value=loss)
    writer.log_metric(key="test_acc", value=accuracy)
    print("loss: {:.2f}".format(loss))
    print("accuracy: {:.2f}".format(accuracy))

    # add artifacts
    writer.log_artifacts(args.output_model_path)
    writer.set_terminated()


if __name__ == "__main__":
    main()



import io
import functools
from pathlib import Path
from typing import Tuple, Optional, Dict

import pandas as pd
import tensorflow as tf
from PIL import Image

from pengu.crawler import ImagesDataCSV
from pengu.dataset.utils import _int64_feature, _bytes_feature


AUTOTUNE = tf.data.experimental.AUTOTUNE


def load_and_resize_image(path: Path,
                          img_size: Optional[Tuple[int, int]] = None,
                          serialize: bool = False):
    """
    Examples:
        if set img_size, use unctools.partial.
        >>> load_and_resize_from_path_fn \
            = functools.partial(load_and_resize_from_path, img_size=(img_h, img_w))
        >>> ds = ds.map(load_and_resize_from_path_fn)
    """
    img = Image.open(str(path))
    if img.mode != "RGB":
        img = img.convert("RGB")

    if img_size is not None:
        img = img.resize((img_size[1], img_size[0]))
    img = tf.image.decode_image(img, channels=3)
    if img_size is not None:
        img = tf.image.resize(img, img_size)
    if serialize:
        img = tf.io.serialize_tensor(img)
    return img


def make_tf_dataset(
        df: pd.DataFrame,
        img_dst_path: Path,
        img_size: Optional[Tuple[int, int]] = None,
        serialize: bool = True) -> tf.data.Dataset:
    image_paths = df[ImagesDataCSV.FILE_PATH_COL].to_list()
    image_paths = [str(img_dst_path / path) for path in image_paths]
    labels = df[ImagesDataCSV.LABEL_COL].to_list()
    ds = tf.data.Dataset.from_tensor_slices((image_paths, labels))
    load_and_resize_from_path_fn \
        = functools.partial(load_and_resize_image, img_size=img_size, serialize=serialize)
    ds = ds.map(load_and_resize_from_path_fn, num_parallel_calls=AUTOTUNE)
    return ds


class ImageTFRecord(object):
    IMG_KEY: str = "image"
    LABEL_KEY: str = "label"
    HEIGHT_KEY: str = "height"
    WIDTH_KEY: str = "width"
    CHANNELS_KEY: str = "channels"

    @classmethod
    def _serialize_example(cls,
                           path: Path,
                           label: int,
                           img_size: Optional[Tuple[int, int]] = None) -> tf.train.Example:
        img = Image.open(str(path))
        if img.mode != "RGB":
            img = img.convert("RGB")

        if img_size is not None:
            img = img.resize((img_size[1], img_size[0]))

        _img_bytes = io.BytesIO()
        img.save(_img_bytes, format="JPEG")
        img_bytes = _img_bytes.getvalue()

        img_w, img_h = img.size
        img_shape: Tuple[int, int, int] = (img_h, img_w, 3)
        feature = {
            cls.LABEL_KEY: _int64_feature(label),
            cls.IMG_KEY: _bytes_feature(img_bytes),
            cls.HEIGHT_KEY: _int64_feature(img_shape[0]),
            cls.WIDTH_KEY: _int64_feature(img_shape[1]),
            cls.CHANNELS_KEY: _int64_feature(img_shape[2]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _desrialize_example(cls, serialized_example):
        image_feature_description = {
            cls.LABEL_KEY: tf.io.FixedLenFeature([], tf.int64),
            cls.IMG_KEY: tf.io.FixedLenFeature([], tf.string),
            cls.HEIGHT_KEY: tf.io.FixedLenFeature((), tf.int64),
            cls.WIDTH_KEY: tf.io.FixedLenFeature((), tf.int64),
            cls.CHANNELS_KEY: tf.io.FixedLenFeature((), tf.int64)
        }
        example = tf.io.parse_single_example(serialized_example,
                                             image_feature_description)
        image = tf.io.decode_image(example[cls.IMG_KEY], channels=3)
        image_shape = [example[cls.HEIGHT_KEY], example[cls.WIDTH_KEY], example[cls.CHANNELS_KEY]]
        image = tf.reshape(image, image_shape)
        return image, example[cls.LABEL_KEY]

    @classmethod
    def write_tfrecord(cls,
                       df: pd.DataFrame,
                       img_dst_path: Path,
                       tfrecord_path: Path,
                       label2index: Optional[Dict[str, int]] = None,
                       img_size: Optional[Tuple[int, int]] = None) -> Dict[str, int]:
        image_paths = df[ImagesDataCSV.FILE_PATH_COL].to_list()
        image_paths = [img_dst_path / path for path in image_paths]
        labels = df[ImagesDataCSV.LABEL_COL].to_list()
        if label2index is None:
            _label2index: Dict[str, int] = {label: i for i, label in enumerate(list(set(labels)))}
        else:
            _label2index = label2index

        with tf.io.TFRecordWriter(str(tfrecord_path)) as writer:
            for path, label in zip(image_paths, labels):
                label_i = _label2index[label]
                tf_example = cls._serialize_example(path, label_i, img_size)
                writer.write(tf_example.SerializeToString())
        return _label2index

    @classmethod
    def read_tfrecord(cls, tfrecord_path: Path) -> tf.data.Dataset:
        raw_ds = tf.data.TFRecordDataset(str(tfrecord_path))

        parsed_ds = raw_ds.map(cls._desrialize_example, num_parallel_calls=AUTOTUNE)
        return parsed_ds



import json
import re
import io
from pathlib import Path
from six.moves.urllib.parse import urlparse
from typing import Dict, Any, List, Tuple

import pandas as pd
from PIL import Image
from bs4 import BeautifulSoup
from icrawler.builtin.google import GoogleParser
from icrawler import ImageDownloader
from sklearn.model_selection import train_test_split

from pengu.data.image_utils import get_image_meta


class ImagesDataCSV(object):
    DELIMITER = ","
    URL_COL: str = "url"
    FILE_PATH_COL: str = "file_path"
    LABEL_COL: str = "label"
    IMG_W_COL: str = "img_w"
    IMG_H_COL: str = "img_h"
    MODE_COL: str = "mode"
    HASH_COL: str = "hash"
    FORMAT_COL: str = "format"
    HEADER: List[str] = [URL_COL, FILE_PATH_COL, LABEL_COL,
                         IMG_W_COL, IMG_H_COL, MODE_COL,
                         HASH_COL, FORMAT_COL]

    def __init__(self, csv_path: Path):
        self.csv_path = csv_path

    @classmethod
    def write_header(cls, csv_path: Path):
        csv_path.parent.mkdir(parents=True, exist_ok=True)
        with open(csv_path, 'w', newline='') as f:
            f.write(cls.DELIMITER.join(cls.HEADER) + "\n")

    def write_row(self,
                  url: str,
                  file_path: str,
                  label: str,
                  img_w: str,
                  img_h: str,
                  mode: str,
                  hash: str,
                  format: str):
        row: List[str] = [url, file_path, label, img_w, img_h, mode, hash, format]
        with open(self.csv_path, 'a', newline='') as f:
            f.write(self.DELIMITER.join(row) + "\n")

    def drop_duplicates_hash(self) -> None:
        df = pd.read_csv(self.csv_path)
        df = df.drop_duplicates(subset=[self.HASH_COL], keep="first")
        self.df = df

    def split_train_val_test(
            self,
            ratio: Tuple[float, float, float]) -> Tuple[pd.DataFrame,
                                                        pd.DataFrame,
                                                        pd.DataFrame]:
        train_ratio, val_ratio, test_ratio = ratio
        df = self.df[[self.FILE_PATH_COL, self.LABEL_COL]]
        labels = df[self.LABEL_COL]
        train_val_df, test_df, _, _ = train_test_split(
            df, labels, test_size=test_ratio, random_state=0, stratify=labels
        )
        train_val_labels = train_val_df[self.LABEL_COL]
        test_size = val_ratio / (train_ratio + val_ratio)
        train_df, val_df, _, _ = train_test_split(
            train_val_df, train_val_labels,
            test_size=test_size, random_state=0, stratify=train_val_labels
        )
        return train_df, val_df, test_df


class SaveResultImageDownloader(ImageDownloader):
    def __init__(self, thread_num, signal, session, storage,
                 image_data_csv_path: Path,
                 label: str,
                 image_exts: List[str] = ["jpeg", "jpg", "png"]):
        """Init Parser with some shared variables."""
        super(SaveResultImageDownloader, self).__init__(
            thread_num, signal, session, storage)

        image_data_csv_path.parent.mkdir(parents=True, exist_ok=True)
        self.image_data_csv = ImagesDataCSV(csv_path=image_data_csv_path)
        self.label = label
        self.image_exts = image_exts

    def process_meta(self, task: Dict[str, Any]):
        if task.get("success"):
            _url = task.get("file_url", "")
            url: str = _url if isinstance(_url, str) else ""
            _file_name = task.get("filename", "")
            file_name: str = _file_name if isinstance(_file_name, str) else ""
            storage_dir: Path = Path(self.storage.root_dir)
            keyword_dir: str = str(storage_dir.name).replace(" ", "_")
            file_path: str = f"{keyword_dir}/{file_name}"
            img_w, img_h = task["img_size"]
            img_mode = task["img_mode"]
            img_hash = task["img_hash"]
            img_format = task["img_format"]

            self.image_data_csv.write_row(
                url=url, file_path=file_path, label=self.label,
                img_w=str(img_w), img_h=str(img_h), mode=img_mode,
                hash=str(img_hash), format=img_format)

    def keep_file(self, task, response, min_size=None, max_size=None):
        """Decide whether to keep the image & add hash
        Compare image size with ``min_size`` and ``max_size`` to decide.

        Args:
            response (Response): response of requests.
            min_size (tuple or None): minimum size of required images.
            max_size (tuple or None): maximum size of required images.

        Returns:
            bool: whether to keep the image.
        """
        try:
            img = Image.open(io.BytesIO(response.content))
        except (IOError, OSError):
            return False
        img_size, img_mode, img_hash, img_format \
            = get_image_meta(img, data_format="pil", hash_method="phash")
        task["img_size"] = img_size
        task["img_mode"] = img_mode
        task["img_hash"] = img_hash
        task["img_format"] = img_format

        if min_size and not self._size_gt(img.size, min_size):
            return False
        if max_size and not self._size_lt(img.size, max_size):
            return False
        return True

    def get_filename(self, task, default_ext):
        url_path = urlparse(task['file_url'])[2]
        if '.' in url_path:
            extension = url_path.split('.')[-1]
            if extension.lower() not in self.image_exts:
                extension = default_ext
        else:
            extension = default_ext
        file_idx = self.fetched_num + self.file_idx_offset
        return '{:06d}.{}'.format(file_idx, extension)

    def worker_exec(self,
                    max_num,
                    default_ext='jpg',
                    queue_timeout=5,
                    req_timeout=5,
                    **kwargs):
        default_req_timeout: int = 1
        default_queue_timeout: int = 1
        super(SaveResultImageDownloader, self).worker_exec(
            max_num, default_ext, queue_timeout=default_queue_timeout,
            req_timeout=default_req_timeout, max_retry=1)


class FixedGoogleParser(GoogleParser):
    def parse(self, response):
        soup = BeautifulSoup(
            response.content.decode('utf-8', 'ignore'), 'lxml')
        image_divs = soup.find_all('script')
        for div in image_divs:
            txt = div.string
            if txt is None or not txt.startswith('AF_initDataCallback'):
                continue
            if 'ds:1' not in txt:
                continue
            # https://github.com/hellock/icrawler/pull/74
            txt = re.sub(r"^AF_initDataCallback\({.*key: 'ds:(\d)'.+data:(.+)}\);?$",
                         "\\2", txt, 0, re.DOTALL)

            meta = json.loads(txt)
            data = meta[31][0][12][2]

            uris = [img[1][3][0] for img in data if img[0] == 1]
            return [{'file_url': uri} for uri in uris]



from pathlib import Path
from dataclasses import dataclass, asdict
from typing import Any, Optional

import tensorflow as tf
from mlflow.tracking.client import MlflowClient

from pengu.utils.file_io import YamlConfig


@dataclass
class MlflowArtifactsConfig(YamlConfig):
    run_id: str
    path: str


@dataclass
class MlflowConfig(YamlConfig):
    experiment_name: str
    tracking_uri: Optional[str] = "http://localhost:5000"
    registry_uri: Optional[str] = None


class MlflowWriter(object):
    def __init__(self,
                 experiment_name: str,
                 tracking_uri: Optional[str] = None,
                 registry_uri: Optional[str] = None):
        self.client = MlflowClient(tracking_uri=tracking_uri, registry_uri=registry_uri)
        try:
            self.experiment_id = self.client.create_experiment(experiment_name)
        except Exception:
            self.experiment_id = self.client.get_experiment_by_name(experiment_name).experiment_id

        self.run_id = self.client.create_run(self.experiment_id).info.run_id

    def log_params_from_config(self, config: object):
        params = asdict(config)
        for key, value in params.items():
            self._log_param_recursive(key, value)

    def _log_param_recursive(self, parent_key, parent_value):
        if isinstance(parent_value, dict):
            for key, value in parent_value.items():
                if isinstance(value, (dict, list, tuple)):
                    self._log_param_recursive(f'{parent_key}.{key}', value)
                else:
                    self.client.log_param(self.run_id, f'{parent_key}.{key}', str(value))
        else:
            self.client.log_param(self.run_id, f'{parent_key}', str(parent_value))

    def log_param(self, key: str, value: Any):
        self.client.log_param(self.run_id, key, str(value))

    def log_metric(self, key: str, value: Any, step: Optional[int] = None):
        self.client.log_metric(self.run_id, key, value, step=step)

    def log_artifact(self, path: Path):
        self.client.log_artifact(self.run_id, str(path))

    def log_artifacts(self, path: Path):
        self.client.log_artifact(self.run_id, str(path))

    def set_terminated(self):
        self.client.set_terminated(self.run_id)

    def download_artifacts(self,
                           run_id: str,
                           path: str,
                           dst_path: Path) -> Path:
        dst_path.mkdir(parents=True, exist_ok=True)
        artifacts_path = self.client.download_artifacts(run_id=run_id,
                                                        path=path,
                                                        dst_path=str(dst_path))
        return Path(artifacts_path)


class MlflowCallback(tf.keras.callbacks.Callback):
    def __init__(self, writer: MlflowWriter):
        self.writer = writer

    def on_epoch_end(self, epoch, logs):
        for key, value in logs.items():
            self.writer.log_metric(key=key, value=value, step=epoch)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pengu',
 'pengu.data',
 'pengu.dataset',
 'pengu.main',
 'pengu.model',
 'pengu.utils']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.14.38,<2.0.0',
 'hydra>=2.5,<3.0',
 'icrawler>=0.6.3,<0.7.0',
 'imagehash>=4.1.0,<5.0.0',
 'matplotlib>=3.3.0,<4.0.0',
 'mlflow>=1.10.0,<2.0.0',
 'mypy>=0.782,<0.783',
 'opencv-python==4.2.0.34',
 'scikit-learn>=0.23.2,<0.24.0',
 'tensorboard_plugin_profile>=2.3.0,<3.0.0',
 'tensorflow==2.3.0']

setup_kwargs = {
    'name': 'pengu',
    'version': '0.1.1',
    'description': 'WIP',
    'long_description': '\n# Pengu: a Library for Deep Learning in Computer Vision\nWork In Progress！\n\n![Test](https://github.com/peachanG/pengu/workflows/Test/badge.svg?branch=master)\n![Python](https://img.shields.io/badge/python-3.7%20%7C%203.8-blue?logo=python)\n[![GitHub Issues](https://img.shields.io/github/issues/peachanG/pengu.svg?cacheSeconds=60&color=yellow)](https://github.com/peachanG/pengu/issues)\n[![GitHub Pull Requests](https://img.shields.io/github/issues-pr/peachanG/pengu.svg?cacheSeconds=60&color=yellow)](https://github.com/peachanG/pengu/issues)\n[![GitHub Release](https://img.shields.io/github/release/peachanG/pengu.svg?cacheSeconds=60&color=red)](https://github.com/peachanG/pengu/releases)\n\n# command\n```\n$ mlflow server --backend-store-uri ./.mlflow_v2 --default-artifact-root s3://pengu-mlflow -p 5001\n$ tensorboard --logdir=./logs\n```\n```\n$ python -m pengu.main.crawler -config config/crawler.yaml -csv data/v2/crawler/crawler_result.csv -i data/v2/crawler/images -n 50\n$ python -m pengu.main.preprocess_images -config config/preprocess.yaml -csv data/preprocessed\n$ python -m pengu.main.train -config config/train.yaml -i data/train/inputs -m data/train/models -t ./logs\n```',
    'author': 'peachanG',
    'author_email': 'kenkman0427@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/peachanG',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.6,<4.0.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Optional, Tuple, Any, Dict

import tensorflow as tf
from tensorflow.keras.applications.resnet50 import ResNet50
from tensorflow.keras.applications.xception import Xception
from tensorflow.keras.applications.vgg16 import VGG16
from tensorflow.keras.applications.vgg19 import VGG19
from tensorflow.keras.applications.inception_v3 import InceptionV3
from tensorflow.keras.applications.inception_resnet_v2 import InceptionResNetV2
from tensorflow.keras.applications.mobilenet import MobileNet
from tensorflow.keras.applications.densenet import DenseNet201
from tensorflow.keras.applications.nasnet import NASNetLarge
from tensorflow.keras.applications.nasnet import NASNetMobile
from tensorflow.keras.applications.mobilenet_v2 import MobileNetV2

from pengu.utils.file_io import YamlConfig


@dataclass
class ModelConfig(YamlConfig):
    name: str
    params: Optional[Dict[str, Any]] = None


def build_imagenet_model(model_name: str,
                         **kwargs) -> tf.keras.Model:
    """build keras application models and Return input shape.

    Args:
        model_name (str):
            Name of model supported by keras.applications.

    Returns:
        keras.engine.training.Model: built model.

    """
    model_name = model_name.lower()
    input_shape = kwargs.pop("input_shape", get_imagenet_input_shape(model_name))

    if model_name == 'xception':
        model = Xception(input_shape=input_shape, **kwargs)
    elif model_name == 'vgg16':
        model = VGG16(input_shape=input_shape, **kwargs)
    elif model_name == 'vgg19':
        model = VGG19(input_shape=input_shape, **kwargs)
    elif model_name == 'resnet50':
        model = ResNet50(input_shape=input_shape, **kwargs)
    elif model_name in ['inception_v3', 'inceptionv3']:
        model = InceptionV3(input_shape=input_shape, **kwargs)
    elif model_name in ['inception_resnet_v2', 'inceptionresnetv2']:
        model = InceptionResNetV2(input_shape=input_shape, **kwargs)
    elif model_name == 'mobilenet':
        model = MobileNet(input_shape=input_shape, **kwargs)
    elif model_name in ['densenet_201', 'densenet201']:
        model = DenseNet201(input_shape=input_shape, **kwargs)
    elif model_name in ['nasnet_large', 'nasnetlarge']:
        model = NASNetLarge(input_shape=input_shape, **kwargs)
    elif model_name in ['nasnet_mobile', 'nasnetmobilne']:
        model = NASNetMobile(input_shape=input_shape, **kwargs)
    elif model_name in ['mobilenet_v2', 'mobilenetv2']:
        model = MobileNetV2(input_shape=input_shape, **kwargs)
    else:
        raise NotImplementedError(f"Not implemented. model_name: {model_name}")

    return model


def get_imagenet_input_shape(model_name: str) -> Optional[Tuple[int, int, int]]:
    """Return input shape of keras.applications model.

    Args:
        model_name (str):
            Name of model supported by keras.applications.

    Returns:
        Optional[Tuple[int, int, int]]: (height, width, channels)

    """
    model_name = model_name.lower()

    if model_name == 'xception':
        input_shape: Optional[Tuple[int, int, int]] = (299, 299, 3)
    elif model_name == 'vgg16':
        input_shape = (224, 224, 3)
    elif model_name == 'vgg19':
        input_shape = (224, 224, 3)
    elif model_name == 'resnet50':
        input_shape = (224, 224, 3)
    elif model_name in ['inception_v3', 'inceptionv3']:
        input_shape = (299, 299, 3)
    elif model_name in ['inception_resnet_v2', 'inceptionresnetv2']:
        input_shape = (299, 299, 3)
    elif model_name == 'mobilenet':
        input_shape = (224, 224, 3)
    elif model_name in ['densenet_201', 'densenet201']:
        input_shape = (224, 224, 3)
    elif model_name in ['nasnet_large', 'nasnetlarge']:
        input_shape = (331, 331, 3)
    elif model_name in ['nasnet_mobile', 'nasnetmobilne']:
        input_shape = (224, 224, 3)
    elif model_name in ['mobilenet_v2', 'mobilenetv2']:
        input_shape = (224, 224, 3)
    else:
        input_shape = None

    return input_shape



import tensorflow as tf


def _bytes_feature(value):
    """string / byte 型から byte_list を返す"""
    if isinstance(value, type(tf.constant(0))):
        # BytesList won't unpack a string from an EagerTensor.
        value = value.numpy()
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _float_feature(value):
    """float / double 型から float_list を返す"""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def _int64_feature(value):
    """bool / enum / int / uint 型から Int64_list を返す"""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))



import re
import time
import logging
import absl.logging
from datetime import datetime
from typing import Any, List

from pytz import timezone, utc


def alpha_numeric_sorted(values: List[str]) -> List[str]:
    """Sort the given iterable in the way that humans expect.
    Args:
        l (List[str]): list will be sorted.
    Returns:
        List[str]: list sorted by alpha numeric method.
    """
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]

    return sorted(values, key=alphanum_key)


def customTime(*args: Any) -> time.struct_time:
    """Convert timezone to "Asia/Tokyo" for logging.
    Returns:
        time.struct_time: Converter timezone to "Asia/Tokyo"
    """
    utc_dt = utc.localize(datetime.utcnow())
    my_tz = timezone("Asia/Tokyo")
    converted = utc_dt.astimezone(my_tz)
    return converted.timetuple()


def set_logging_basic_config(log_level: str = "INFO") -> None:
    """Build logging logger
    Args:
        log_level (str): log_level.
                         Defaults to "INFO".
    """
    # NOTE: https://github.com/tensorflow/tensorflow/issues/27045
    logging.root.removeHandler(absl.logging._absl_handler)
    absl.logging._warn_preinit_stderr = False

    logging.Formatter.converter = customTime
    log_fmt = logging.Formatter('%(asctime)s %(name)s %(lineno)d '
                                '[%(levelname)s][%(funcName)s] %(message)s')
    s_handler = logging.StreamHandler()
    s_handler.setLevel(log_level)
    s_handler.setFormatter(log_fmt)
    logging.basicConfig(handlers=[s_handler], level=log_level)
